@@ -1,36 +1,50 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 
-from __future__ import absolute_import, print_function
+from __future__ import annotations
 
 import ast
 import collections
 import os
 import sys
 
+import typing
+
+
 # Import a minimal set of stdlib modules needed for list_stdlib_modules()
 # to work when run from a virtualenv.  The modules were chosen empirically
 # so that the return value matches the return value without virtualenv.
-if True: # disable lexical sorting checks
-    try:
-        import BaseHTTPServer as basehttpserver
-    except ImportError:
-        basehttpserver = None
+if True:  # disable lexical sorting checks
+    import argparse
     import zlib
 
 import testparseutil
 
-# Whitelist of modules that symbols can be directly imported from.
+if typing.TYPE_CHECKING:
+    from typing import Iterator
+
+    ImportDictT = dict[str, list[str]]
+    # source code, modename, filename, line (embedded-only)
+    SourceT = tuple[bytes, str, str, int]
+
+# Allow list of modules that symbols can be directly imported from.
 allowsymbolimports = (
     '__future__',
-    'bzrlib',
+    'breezy',
+    'concurrent',
     'hgclient',
+    'packaging.version',
     'mercurial',
     'mercurial.hgweb.common',
     'mercurial.hgweb.request',
     'mercurial.i18n',
+    'mercurial.interfaces',
+    'mercurial.interfaces._basetypes',
+    'mercurial.interfaces.types',
     'mercurial.node',
+    'mercurial.pycompat',
     # for revlog to re-export constant to extensions
     'mercurial.revlogutils.constants',
+    'mercurial.revlogutils.flagutil',
     # for cffi modules to re-export pure functions
     'mercurial.pure.base85',
     'mercurial.pure.bdiff',
@@ -40,14 +54,14 @@ allowsymbolimports = (
     # third-party imports should be directly imported
     'mercurial.thirdparty',
     'mercurial.thirdparty.attr',
-    'mercurial.thirdparty.zope',
-    'mercurial.thirdparty.zope.interface',
+    'mercurial.thirdparty.jaraco.collections',
+    'mercurial.thirdparty.tomli',
+    'typing',
+    'xml.etree.ElementTree',
 )
 
-# Whitelist of symbols that can be directly imported.
-directsymbols = (
-    'demandimport',
-)
+# Allow list of symbols that can be directly imported.
+directsymbols = ('demandimport',)
 
 # Modules that must be aliased because they are commonly confused with
 # common variables and can create aliasing and readability issues.
@@ -55,19 +69,6 @@ requirealias = {
     'ui': 'uimod',
 }
 
-def usingabsolute(root):
-    """Whether absolute imports are being used."""
-    if sys.version_info[0] >= 3:
-        return True
-
-    for node in ast.walk(root):
-        if isinstance(node, ast.ImportFrom):
-            if node.module == '__future__':
-                for n in node.names:
-                    if n.name == 'absolute_import':
-                        return True
-
-    return False
 
 def walklocal(root):
     """Recursively yield all descendant nodes but not in a different scope"""
@@ -80,7 +81,8 @@ def walklocal(root):
             todo.extend(ast.iter_child_nodes(node))
         yield node, newscope
 
-def dotted_name_of_path(path):
+
+def dotted_name_of_path(path: str) -> str:
     """Given a relative path to a source file, return its dotted module name.
 
     >>> dotted_name_of_path('mercurial/error.py')
@@ -89,11 +91,12 @@ def dotted_name_of_path(path):
     'zlib'
     """
     parts = path.replace(os.sep, '/').split('/')
-    parts[-1] = parts[-1].split('.', 1)[0] # remove .py and .so and .ARCH.so
+    parts[-1] = parts[-1].split('.', 1)[0]  # remove .py and .so and .ARCH.so
     if parts[-1].endswith('module'):
         parts[-1] = parts[-1][:-6]
     return '.'.join(parts)
 
+
 def fromlocalfunc(modulename, localmods):
     """Get a function to examine which locally defined module the
     target source imports via a specified name.
@@ -162,6 +165,7 @@ def fromlocalfunc(modulename, localmods):
     prefix = '.'.join(modulename.split('.')[:-1])
     if prefix:
         prefix += '.'
+
     def fromlocal(name, level=0):
         # name is false value when relative imports are used.
         if not name:
@@ -173,8 +177,9 @@ def fromlocalfunc(modulename, localmods):
                 # Check relative name first.
                 candidates = [prefix + name, name]
             else:
-                candidates = ['.'.join(modulename.split('.')[:-level]) +
-                              '.' + name]
+                candidates = [
+                    '.'.join(modulename.split('.')[:-level]) + '.' + name
+                ]
 
         for n in candidates:
             if n in localmods:
@@ -183,24 +188,26 @@ def fromlocalfunc(modulename, localmods):
             if dottedpath in localmods:
                 return (n, dottedpath, True)
         return False
+
     return fromlocal
 
+
 def populateextmods(localmods):
     """Populate C extension modules based on pure modules"""
     newlocalmods = set(localmods)
     for n in localmods:
         if n.startswith('mercurial.pure.'):
-            m = n[len('mercurial.pure.'):]
+            m = n[len('mercurial.pure.') :]
             newlocalmods.add('mercurial.cext.' + m)
             newlocalmods.add('mercurial.cffi._' + m)
     return newlocalmods
 
+
 def list_stdlib_modules():
     """List the modules present in the stdlib.
 
-    >>> py3 = sys.version_info[0] >= 3
     >>> mods = set(list_stdlib_modules())
-    >>> 'BaseHTTPServer' in mods or py3
+    >>> 'http' in mods
     True
 
     os.path isn't really a module, so it's missing:
@@ -217,33 +224,29 @@ def list_stdlib_modules():
     >>> 'collections' in mods
     True
 
-    >>> 'cStringIO' in mods or py3
+    >>> 'array' in mods
     True
 
     >>> 'cffi' in mods
     True
     """
-    for m in sys.builtin_module_names:
-        yield m
+    yield from sys.builtin_module_names
     # These modules only exist on windows, but we should always
     # consider them stdlib.
-    for m in ['msvcrt', '_winreg']:
-        yield m
+    yield from ['msvcrt', '_winreg']
     yield '__builtin__'
-    yield 'builtins' # python3 only
-    yield 'importlib.abc' # python3 only
-    yield 'importlib.machinery' # python3 only
-    yield 'importlib.util' # python3 only
-    for m in 'fcntl', 'grp', 'pwd', 'termios':  # Unix only
-        yield m
-    for m in 'cPickle', 'datetime': # in Python (not C) on PyPy
-        yield m
-    for m in ['cffi']:
-        yield m
+    yield 'builtins'  # python3 only
+    yield 'importlib.abc'  # python3 only
+    yield 'importlib.machinery'  # python3 only
+    yield 'importlib.util'  # python3 only
+    yield 'packaging.version'
+    yield from ['fcntl', 'grp', 'pwd', 'select', 'termios']
+    yield from ['cPickle', 'datetime']
+    yield from ['cffi']
     stdlib_prefixes = {sys.prefix, sys.exec_prefix}
     # We need to supplement the list of prefixes for the search to work
     # when run from within a virtualenv.
-    for mod in (basehttpserver, zlib):
+    for mod in (argparse, zlib):
         if mod is None:
             continue
         try:
@@ -262,14 +265,19 @@ def list_stdlib_modules():
     for libpath in sys.path:
         # We want to walk everything in sys.path that starts with something in
         # stdlib_prefixes, but not directories from the hg sources.
-        if (os.path.abspath(libpath).startswith(sourceroot)
-            or not any(libpath.startswith(p) for p in stdlib_prefixes)):
+        if os.path.abspath(libpath).startswith(sourceroot) or not any(
+            libpath.startswith(p) for p in stdlib_prefixes
+        ):
             continue
         for top, dirs, files in os.walk(libpath):
+            if 'dist-packages' in top.split(os.path.sep):
+                continue
             for i, d in reversed(list(enumerate(dirs))):
-                if (not os.path.exists(os.path.join(top, d, '__init__.py'))
-                    or top == libpath and d in ('hgdemandimport', 'hgext',
-                                                'mercurial')):
+                if (
+                    not os.path.exists(os.path.join(top, d, '__init__.py'))
+                    or top == libpath
+                    and d in ('hgdemandimport', 'hgext', 'mercurial')
+                ):
                     del dirs[i]
             for name in files:
                 if not name.endswith(('.py', '.so', '.pyc', '.pyo', '.pyd')):
@@ -278,12 +286,14 @@ def list_stdlib_modules():
                     full_path = top
                 else:
                     full_path = os.path.join(top, name)
-                rel_path = full_path[len(libpath) + 1:]
+                rel_path = full_path[len(libpath) + 1 :]
                 mod = dotted_name_of_path(rel_path)
                 yield mod
 
+
 stdlib_modules = set(list_stdlib_modules())
 
+
 def imported_modules(source, modulename, f, localmods, ignore_nested=False):
     """Given the source of a file as a string, yield the names
     imported by that file.
@@ -376,27 +386,19 @@ def imported_modules(source, modulename, f, localmods, ignore_nested=False):
                     modnotfound = True
                     continue
                 yield found[1]
-            if modnotfound:
+            if modnotfound and dottedpath != modulename:
                 # "dottedpath" is a package, but imported because of non-module
                 # lookup
+                # specifically allow "from . import foo" from __init__.py
                 yield dottedpath
 
+
 def verify_import_convention(module, source, localmods):
-    """Verify imports match our established coding convention.
-
-    We have 2 conventions: legacy and modern. The modern convention is in
-    effect when using absolute imports.
-
-    The legacy convention only looks for mixed imports. The modern convention
-    is much more thorough.
-    """
+    """Verify imports match our established coding convention."""
     root = ast.parse(source)
-    absolute = usingabsolute(root)
 
-    if absolute:
-        return verify_modern_convention(module, root, localmods)
-    else:
-        return verify_stdlib_on_own_line(root)
+    return verify_modern_convention(module, root, localmods)
+
 
 def verify_modern_convention(module, root, localmods, root_col_offset=0):
     """Verify a file conforms to the modern import convention rules.
@@ -441,19 +443,23 @@ def verify_modern_convention(module, root, localmods, root_col_offset=0):
     seenlevels = set()
 
     for node, newscope in walklocal(root):
+
         def msg(fmt, *args):
             return (fmt % args, node.lineno)
+
         if newscope:
             # Check for local imports in function
-            for r in verify_modern_convention(module, node, localmods,
-                                              node.col_offset + 4):
-                yield r
+            yield from verify_modern_convention(
+                module, node, localmods, node.col_offset + 4
+            )
         elif isinstance(node, ast.Import):
             # Disallow "import foo, bar" and require separate imports
             # for each module.
             if len(node.names) > 1:
-                yield msg('multiple imported names: %s',
-                          ', '.join(n.name for n in node.names))
+                yield msg(
+                    'multiple imported names: %s',
+                    ', '.join(n.name for n in node.names),
+                )
 
             name = node.names[0].name
             asname = node.names[0].asname
@@ -463,16 +469,20 @@ def verify_modern_convention(module, root, localmods, root_col_offset=0):
             # Ignore sorting rules on imports inside blocks.
             if node.col_offset == root_col_offset:
                 if lastname and name < lastname and laststdlib == stdlib:
-                    yield msg('imports not lexically sorted: %s < %s',
-                              name, lastname)
+                    yield msg(
+                        'imports not lexically sorted: %s < %s', name, lastname
+                    )
 
             lastname = name
             laststdlib = stdlib
 
             # stdlib imports should be before local imports.
             if stdlib and seenlocal and node.col_offset == root_col_offset:
-                yield msg('stdlib import "%s" follows local import: %s',
-                          name, seenlocal)
+                yield msg(
+                    'stdlib import "%s" follows local import: %s',
+                    name,
+                    seenlocal,
+                )
 
             if not stdlib:
                 seenlocal = name
@@ -483,13 +493,16 @@ def verify_modern_convention(module, root, localmods, root_col_offset=0):
                 yield msg('import should be relative: %s', name)
 
             if name in requirealias and asname != requirealias[name]:
-                yield msg('%s module must be "as" aliased to %s',
-                          name, requirealias[name])
+                yield msg(
+                    '%s module must be "as" aliased to %s',
+                    name,
+                    requirealias[name],
+                )
 
         elif isinstance(node, ast.ImportFrom):
             # Resolve the full imported module name.
             if node.level > 0:
-                fullname = '.'.join(module.split('.')[:-node.level])
+                fullname = '.'.join(module.split('.')[: -node.level])
                 if node.module:
                     fullname += '.%s' % node.module
             else:
@@ -505,8 +518,11 @@ def verify_modern_convention(module, root, localmods, root_col_offset=0):
             if fullname != '__future__':
                 if not fullname or (
                     fullname in stdlib_modules
+                    # allow standard 'from typing import ...' style
+                    and fullname.startswith('.')
                     and fullname not in localmods
-                    and fullname + '.__init__' not in localmods):
+                    and fullname + '.__init__' not in localmods
+                ):
                     yield msg('relative import of stdlib module')
                 else:
                     seenlocal = fullname
@@ -516,19 +532,24 @@ def verify_modern_convention(module, root, localmods, root_col_offset=0):
             found = fromlocal(node.module, node.level)
             if found and found[2]:  # node.module is a package
                 prefix = found[0] + '.'
-                symbols = (n.name for n in node.names
-                           if not fromlocal(prefix + n.name))
+                symbols = (
+                    n.name for n in node.names if not fromlocal(prefix + n.name)
+                )
             else:
                 symbols = (n.name for n in node.names)
             symbols = [sym for sym in symbols if sym not in directsymbols]
             if node.module and node.col_offset == root_col_offset:
                 if symbols and fullname not in allowsymbolimports:
-                    yield msg('direct symbol import %s from %s',
-                              ', '.join(symbols), fullname)
+                    yield msg(
+                        'direct symbol import %s from %s',
+                        ', '.join(symbols),
+                        fullname,
+                    )
 
                 if symbols and seennonsymbollocal:
-                    yield msg('symbol import follows non-symbol import: %s',
-                              fullname)
+                    yield msg(
+                        'symbol import follows non-symbol import: %s', fullname
+                    )
             if not symbols and fullname not in stdlib_modules:
                 seennonsymbollocal = True
 
@@ -536,15 +557,19 @@ def verify_modern_convention(module, root, localmods, root_col_offset=0):
                 assert node.level
 
                 # Only allow 1 group per level.
-                if (node.level in seenlevels
-                    and node.col_offset == root_col_offset):
-                    yield msg('multiple "from %s import" statements',
-                              '.' * node.level)
+                if (
+                    node.level in seenlevels
+                    and node.col_offset == root_col_offset
+                ):
+                    yield msg(
+                        'multiple "from %s import" statements', '.' * node.level
+                    )
 
                 # Higher-level groups come before lower-level groups.
                 if any(node.level > l for l in seenlevels):
-                    yield msg('higher-level import should come first: %s',
-                              fullname)
+                    yield msg(
+                        'higher-level import should come first: %s', fullname
+                    )
 
                 seenlevels.add(node.level)
 
@@ -554,42 +579,31 @@ def verify_modern_convention(module, root, localmods, root_col_offset=0):
 
             for n in node.names:
                 if lastentryname and n.name < lastentryname:
-                    yield msg('imports from %s not lexically sorted: %s < %s',
-                              fullname, n.name, lastentryname)
+                    yield msg(
+                        'imports from %s not lexically sorted: %s < %s',
+                        fullname,
+                        n.name,
+                        lastentryname,
+                    )
 
                 lastentryname = n.name
 
                 if n.name in requirealias and n.asname != requirealias[n.name]:
-                    yield msg('%s from %s must be "as" aliased to %s',
-                              n.name, fullname, requirealias[n.name])
+                    yield msg(
+                        '%s from %s must be "as" aliased to %s',
+                        n.name,
+                        fullname,
+                        requirealias[n.name],
+                    )
 
-def verify_stdlib_on_own_line(root):
-    """Given some python source, verify that stdlib imports are done
-    in separate statements from relative local module imports.
-
-    >>> list(verify_stdlib_on_own_line(ast.parse('import sys, foo')))
-    [('mixed imports\\n   stdlib:    sys\\n   relative:  foo', 1)]
-    >>> list(verify_stdlib_on_own_line(ast.parse('import sys, os')))
-    []
-    >>> list(verify_stdlib_on_own_line(ast.parse('import foo, bar')))
-    []
-    """
-    for node in ast.walk(root):
-        if isinstance(node, ast.Import):
-            from_stdlib = {False: [], True: []}
-            for n in node.names:
-                from_stdlib[n.name in stdlib_modules].append(n.name)
-            if from_stdlib[True] and from_stdlib[False]:
-                yield ('mixed imports\n   stdlib:    %s\n   relative:  %s' %
-                       (', '.join(sorted(from_stdlib[True])),
-                        ', '.join(sorted(from_stdlib[False]))), node.lineno)
 
 class CircularImport(Exception):
     pass
 
-def checkmod(mod, imports):
-    shortest = {}
-    visit = [[mod]]
+
+def checkmod(mod: str, imports: ImportDictT) -> None:
+    shortest: dict[str, int] = {}
+    visit: list[list[str]] = [[mod]]
     while visit:
         path = visit.pop(0)
         for i in sorted(imports.get(path[-1], [])):
@@ -601,7 +615,8 @@ def checkmod(mod, imports):
                     continue
                 visit.append(path + [i])
 
-def rotatecycle(cycle):
+
+def rotatecycle(cycle: list[str]) -> list[str]:
     """arrange a cycle so that the lexicographically first module listed first
 
     >>> rotatecycle(['foo', 'bar'])
@@ -611,12 +626,12 @@ def rotatecycle(cycle):
     idx = cycle.index(lowest)
     return cycle[idx:] + cycle[:idx] + [lowest]
 
-def find_cycles(imports):
+
+def find_cycles(imports: ImportDictT) -> set[str]:
     """Find cycles in an already-loaded import graph.
 
     All module names recorded in `imports` should be absolute one.
 
-    >>> from __future__ import print_function
     >>> imports = {'top.foo': ['top.bar', 'os.path', 'top.qux'],
     ...            'top.bar': ['top.baz', 'sys'],
     ...            'top.baz': ['top.foo'],
@@ -630,14 +645,16 @@ def find_cycles(imports):
         try:
             checkmod(mod, imports)
         except CircularImport as e:
-            cycle = e.args[0]
+            cycle: list[str] = e.args[0]
             cycles.add(" -> ".join(rotatecycle(cycle)))
     return cycles
 
+
 def _cycle_sortkey(c):
     return len(c), c
 
-def embedded(f, modname, src):
+
+def embedded(f, modname, src) -> Iterator[SourceT]:
     """Extract embedded python code
 
     >>> def _forcestr(thing):
@@ -649,15 +666,15 @@ def embedded(f, modname, src):
     ...         print("%s %s %d" % (_forcestr(m), _forcestr(f), l))
     ...         print(repr(_forcestr(s)))
     >>> lines = [
-    ...   b'comment',
-    ...   b'  >>> from __future__ import print_function',
-    ...   b"  >>> ' multiline",
-    ...   b"  ... string'",
-    ...   b'  ',
-    ...   b'comment',
-    ...   b'  $ cat > foo.py <<EOF',
-    ...   b'  > from __future__ import print_function',
-    ...   b'  > EOF',
+    ...   'comment',
+    ...   '  >>> from __future__ import print_function',
+    ...   "  >>> ' multiline",
+    ...   "  ... string'",
+    ...   '  ',
+    ...   'comment',
+    ...   '  $ cat > foo.py <<EOF',
+    ...   '  > from __future__ import print_function',
+    ...   '  > EOF',
     ... ]
     >>> test(b"example.t", lines)
     example[2] doctest.py 1
@@ -665,7 +682,7 @@ def embedded(f, modname, src):
     example[8] foo.py 7
     'from __future__ import print_function\\n'
     """
-    errors = []
+    errors: list[str] = []
     for name, starts, ends, code in testparseutil.pyembedded(f, src, errors):
         if not name:
             # use 'doctest.py', in order to make already existing
@@ -678,7 +695,8 @@ def embedded(f, modname, src):
             modname = modname.decode('utf8')
         yield code, "%s[%d]" % (modname, starts), name, starts - 1
 
-def sources(f, modname):
+
+def sources(f: str, modname: str) -> Iterator[SourceT]:
     """Yields possibly multiple sources from a filepath
 
     input: filepath, modulename
@@ -694,11 +712,15 @@ def sources(f, modname):
             yield src.read(), modname, f, 0
             py = True
     if py or f.endswith('.t'):
-        with open(f, 'rb') as src:
-            for script, modname, t, line in embedded(f, modname, src):
-                yield script, modname.encode('utf8'), t, line
+        # Strictly speaking we should sniff for the magic header that denotes
+        # Python source file encoding. But in reality we don't use anything
+        # other than ASCII (mainly) and UTF-8 (in a few exceptions), so
+        # simplicity is fine.
+        with open(f, encoding='utf-8') as src:
+            yield from embedded(f, modname, src)
 
-def main(argv):
+
+def main(argv: list[str]) -> int:
     if len(argv) < 2 or (argv[1] == '-' and len(argv) > 2):
         print('Usage: %s {-|file [file] [file] ...}')
         return 1
@@ -706,7 +728,7 @@ def main(argv):
         argv = argv[:1]
         argv.extend(l.rstrip() for l in sys.stdin.readlines())
     localmodpaths = {}
-    used_imports = {}
+    used_imports: ImportDictT = {}
     any_errors = False
     for source_path in argv[1:]:
         modname = dotted_name_of_path(source_path)
@@ -715,19 +737,24 @@ def main(argv):
     for localmodname, source_path in sorted(localmodpaths.items()):
         if not isinstance(localmodname, bytes):
             # This is only safe because all hg's files are ascii
-            localmodname = localmodname.encode('ascii')
+            localmodname = localmodname
         for src, modname, name, line in sources(source_path, localmodname):
             try:
                 used_imports[modname] = sorted(
-                    imported_modules(src, modname, name, localmods,
-                                     ignore_nested=True))
-                for error, lineno in verify_import_convention(modname, src,
-                                                              localmods):
+                    imported_modules(
+                        src, modname, name, localmods, ignore_nested=True
+                    )
+                )
+                for error, lineno in verify_import_convention(
+                    modname, src, localmods
+                ):
                     any_errors = True
                     print('%s:%d: %s' % (source_path, lineno + line, error))
             except SyntaxError as e:
-                print('%s:%d: SyntaxError: %s' %
-                      (source_path, e.lineno + line, e))
+                assert e.lineno is not None
+                print(
+                    '%s:%d: SyntaxError: %s' % (source_path, e.lineno + line, e)
+                )
     cycles = find_cycles(used_imports)
     if cycles:
         firstmods = set()
@@ -743,5 +770,6 @@ def main(argv):
         any_errors = True
     return any_errors != 0
 
+
 if __name__ == '__main__':
     sys.exit(int(main(sys.argv)))
