@@ -5,7 +5,7 @@
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
-from __future__ import absolute_import, print_function
+from __future__ import annotations
 
 import errno
 import getopt
@@ -17,9 +17,11 @@ import signal
 import sys
 import traceback
 
+from typing import (
+    Iterable,
+)
 
 from .i18n import _
-from .pycompat import getattr
 
 from hgdemandimport import tracing
 
@@ -27,6 +29,7 @@ from . import (
     cmdutil,
     color,
     commands,
+    config as configmod,
     demandimport,
     encoding,
     error,
@@ -35,18 +38,15 @@ from . import (
     help,
     hg,
     hook,
-    localrepo,
     profiling,
     pycompat,
-    rcutil,
     registrar,
-    requirements as requirementsmod,
     scmutil,
     ui as uimod,
     util,
-    vfs,
 )
 
+from .configuration import rcutil
 from .utils import (
     procutil,
     stringutil,
@@ -54,7 +54,7 @@ from .utils import (
 )
 
 
-class request(object):
+class request:
     def __init__(
         self,
         args,
@@ -87,7 +87,7 @@ class request(object):
         # store the parsed and canonical command
         self.canonical_command = None
 
-    def _runexithandlers(self):
+    def _runexithandlers(self) -> None:
         exc = None
         handlers = self.ui._exithandlers
         try:
@@ -108,16 +108,16 @@ class request(object):
 def _flushstdio(ui, err):
     status = None
     # In all cases we try to flush stdio streams.
-    if util.safehasattr(ui, b'fout'):
+    if hasattr(ui, 'fout'):
         assert ui is not None  # help pytype
         assert ui.fout is not None  # help pytype
         try:
             ui.fout.flush()
-        except IOError as e:
+        except OSError as e:
             err = e
             status = -1
 
-    if util.safehasattr(ui, b'ferr'):
+    if hasattr(ui, 'ferr'):
         assert ui is not None  # help pytype
         assert ui.ferr is not None  # help pytype
         try:
@@ -128,7 +128,7 @@ def _flushstdio(ui, err):
             ui.ferr.flush()
         # There's not much we can do about an I/O error here. So (possibly)
         # change the status code and move on.
-        except IOError:
+        except OSError:
             status = -1
 
     return status
@@ -150,93 +150,76 @@ def run():
     sys.exit(status & 255)
 
 
-if pycompat.ispy3:
+def initstdio():
+    # stdio streams on Python 3 are io.TextIOWrapper instances proxying another
+    # buffer. These streams will normalize \n to \r\n by default. Mercurial's
+    # preferred mechanism for writing output (ui.write()) uses io.BufferedWriter
+    # instances, which write to the underlying stdio file descriptor in binary
+    # mode. ui.write() uses \n for line endings and no line ending normalization
+    # is attempted through this interface. This "just works," even if the system
+    # preferred line ending is not \n.
+    #
+    # But some parts of Mercurial (e.g. hooks) can still send data to sys.stdout
+    # and sys.stderr. They will inherit the line ending normalization settings,
+    # potentially causing e.g. \r\n to be emitted. Since emitting \n should
+    # "just work," here we change the sys.* streams to disable line ending
+    # normalization, ensuring compatibility with our ui type.
 
-    def initstdio():
-        # stdio streams on Python 3 are io.TextIOWrapper instances proxying another
-        # buffer. These streams will normalize \n to \r\n by default. Mercurial's
-        # preferred mechanism for writing output (ui.write()) uses io.BufferedWriter
-        # instances, which write to the underlying stdio file descriptor in binary
-        # mode. ui.write() uses \n for line endings and no line ending normalization
-        # is attempted through this interface. This "just works," even if the system
-        # preferred line ending is not \n.
-        #
-        # But some parts of Mercurial (e.g. hooks) can still send data to sys.stdout
-        # and sys.stderr. They will inherit the line ending normalization settings,
-        # potentially causing e.g. \r\n to be emitted. Since emitting \n should
-        # "just work," here we change the sys.* streams to disable line ending
-        # normalization, ensuring compatibility with our ui type.
+    if sys.stdout is not None:
+        # write_through is new in Python 3.7.
+        kwargs = {
+            "newline": "\n",
+            "line_buffering": sys.stdout.line_buffering,
+        }
+        if hasattr(sys.stdout, "write_through"):
+            # pytype: disable=attribute-error
+            kwargs["write_through"] = sys.stdout.write_through
+            # pytype: enable=attribute-error
+        sys.stdout = io.TextIOWrapper(
+            sys.stdout.buffer, sys.stdout.encoding, sys.stdout.errors, **kwargs
+        )
 
-        if sys.stdout is not None:
-            # write_through is new in Python 3.7.
-            kwargs = {
-                "newline": "\n",
-                "line_buffering": sys.stdout.line_buffering,
-            }
-            if util.safehasattr(sys.stdout, "write_through"):
-                # pytype: disable=attribute-error
-                kwargs["write_through"] = sys.stdout.write_through
-                # pytype: enable=attribute-error
-            sys.stdout = io.TextIOWrapper(
-                sys.stdout.buffer,
-                sys.stdout.encoding,
-                sys.stdout.errors,
-                **kwargs
-            )
+    if sys.stderr is not None:
+        kwargs = {
+            "newline": "\n",
+            "line_buffering": sys.stderr.line_buffering,
+        }
+        if hasattr(sys.stderr, "write_through"):
+            # pytype: disable=attribute-error
+            kwargs["write_through"] = sys.stderr.write_through
+            # pytype: enable=attribute-error
+        sys.stderr = io.TextIOWrapper(
+            sys.stderr.buffer, sys.stderr.encoding, sys.stderr.errors, **kwargs
+        )
 
-        if sys.stderr is not None:
-            kwargs = {
-                "newline": "\n",
-                "line_buffering": sys.stderr.line_buffering,
-            }
-            if util.safehasattr(sys.stderr, "write_through"):
-                # pytype: disable=attribute-error
-                kwargs["write_through"] = sys.stderr.write_through
-                # pytype: enable=attribute-error
-            sys.stderr = io.TextIOWrapper(
-                sys.stderr.buffer,
-                sys.stderr.encoding,
-                sys.stderr.errors,
-                **kwargs
-            )
-
-        if sys.stdin is not None:
-            # No write_through on read-only stream.
-            sys.stdin = io.TextIOWrapper(
-                sys.stdin.buffer,
-                sys.stdin.encoding,
-                sys.stdin.errors,
-                # None is universal newlines mode.
-                newline=None,
-                line_buffering=sys.stdin.line_buffering,
-            )
-
-    def _silencestdio():
-        for fp in (sys.stdout, sys.stderr):
-            if fp is None:
-                continue
-            # Check if the file is okay
-            try:
-                fp.flush()
-                continue
-            except IOError:
-                pass
-            # Otherwise mark it as closed to silence "Exception ignored in"
-            # message emitted by the interpreter finalizer.
-            try:
-                fp.close()
-            except IOError:
-                pass
+    if sys.stdin is not None:
+        # No write_through on read-only stream.
+        sys.stdin = io.TextIOWrapper(
+            sys.stdin.buffer,
+            sys.stdin.encoding,
+            sys.stdin.errors,
+            # None is universal newlines mode.
+            newline=None,
+            line_buffering=sys.stdin.line_buffering,
+        )
 
 
-else:
-
-    def initstdio():
-        for fp in (sys.stdin, sys.stdout, sys.stderr):
-            procutil.setbinary(fp)
-
-    def _silencestdio():
-        pass
+def _silencestdio():
+    for fp in (sys.stdout, sys.stderr):
+        if fp is None:
+            continue
+        # Check if the file is okay
+        try:
+            fp.flush()
+            continue
+        except OSError:
+            pass
+        # Otherwise mark it as closed to silence "Exception ignored in"
+        # message emitted by the interpreter finalizer.
+        try:
+            fp.close()
+        except OSError:
+            pass
 
 
 def _formatargs(args):
@@ -258,7 +241,7 @@ def dispatch(req):
     return status
 
 
-def _rundispatch(req):
+def _rundispatch(req) -> int:
     with tracing.log('dispatch._rundispatch'):
         if req.ferr:
             ferr = req.ferr
@@ -273,6 +256,7 @@ def _rundispatch(req):
             req.earlyoptions.update(_earlyparseopts(req.ui, req.args))
             if req.earlyoptions[b'traceback']:
                 req.ui.setconfig(b'ui', b'traceback', b'on', b'--traceback')
+                encoding.enable_rust_backtrace()
 
             # set ui streams from the request
             if req.fin:
@@ -287,7 +271,7 @@ def _rundispatch(req):
             ferr.write(inst.format())
             return -1
 
-        msg = _formatargs(req.args)
+        formattedargs = _formatargs(req.args)
         starttime = util.timer()
         ret = 1  # default of Python exit code on unhandled exception
         try:
@@ -308,9 +292,8 @@ def _rundispatch(req):
                 # maybe pager would quit without consuming all the output, and
                 # SIGPIPE was raised. we cannot print anything in this case.
                 pass
-            except IOError as inst:
-                if inst.errno != errno.EPIPE:
-                    raise
+            except BrokenPipeError:
+                pass
             ret = -1
         finally:
             duration = util.timer() - starttime
@@ -320,13 +303,13 @@ def _rundispatch(req):
                 req.ui.log(
                     b'uiblocked',
                     b'ui blocked ms\n',
-                    **pycompat.strkwargs(req.ui._blockedtimes)
+                    **pycompat.strkwargs(req.ui._blockedtimes),
                 )
             return_code = ret & 255
             req.ui.log(
                 b"commandfinish",
                 b"%s exited %d after %0.2f seconds\n",
-                msg,
+                formattedargs,
                 return_code,
                 duration,
                 return_code=return_code,
@@ -350,7 +333,7 @@ def _runcatch(req):
 
         ui = req.ui
         try:
-            for name in b'SIGBREAK', b'SIGHUP', b'SIGTERM':
+            for name in 'SIGBREAK', 'SIGHUP', 'SIGTERM':
                 num = getattr(signal, name, None)
                 if num:
                     signal.signal(num, catchterm)
@@ -386,12 +369,18 @@ def _runcatch(req):
                 # shenanigans wherein a user does something like pass
                 # --debugger or --config=ui.debugger=1 as a repo
                 # name. This used to actually run the debugger.
+                nbargs = 4
+                hashiddenaccess = b'--hidden' in cmdargs
+                if hashiddenaccess:
+                    nbargs += 1
                 if (
-                    len(req.args) != 4
+                    len(req.args) != nbargs
                     or req.args[0] != b'-R'
                     or req.args[1].startswith(b'--')
                     or req.args[2] != b'serve'
                     or req.args[3] != b'--stdio'
+                    or hashiddenaccess
+                    and req.args[4] != b'--hidden'
                 ):
                     raise error.Abort(
                         _(b'potentially unsafe serve --stdio invocation: %s')
@@ -403,13 +392,22 @@ def _runcatch(req):
                 debugtrace = {b'pdb': pdb.set_trace}
                 debugmortem = {b'pdb': pdb.post_mortem}
 
-                # read --config before doing anything else
-                # (e.g. to change trust settings for reading .hg/hgrc)
+                # read --config-file and --config before doing anything else
+                # (e.g. to change trust settings for reading .hg/hgrc).
+
+                # cmdargs may not have been initialized here (in the case of an
+                # error), so use pycompat.sysargv instead.
+                file_cfgs = _parse_config_files(
+                    req.ui, pycompat.sysargv, req.earlyoptions[b'config_file']
+                )
                 cfgs = _parseconfig(req.ui, req.earlyoptions[b'config'])
 
                 if req.repo:
                     # copy configs that were passed on the cmdline (--config) to
                     # the repo ui
+                    for sec, name, val, source in file_cfgs:
+                        req.repo.ui.setconfig(sec, name, val, source=source)
+
                     for sec, name, val in cfgs:
                         req.repo.ui.setconfig(
                             sec, name, val, source=b'--config'
@@ -427,7 +425,7 @@ def _runcatch(req):
                     # debugging has been requested
                     with demandimport.deactivated():
                         try:
-                            debugmod = __import__(debugger)
+                            debugmod = __import__(pycompat.sysstr(debugger))
                         except ImportError:
                             pass  # Leave debugmod = pdb
 
@@ -516,7 +514,7 @@ def _callcatch(ui, func):
             if not suggested:
                 ui.warn(nocmdmsg)
                 ui.warn(_(b"(use 'hg help' for a list of commands)\n"))
-    except IOError:
+    except OSError:
         raise
     except KeyboardInterrupt:
         raise
@@ -533,7 +531,7 @@ def _callcatch(ui, func):
 def aliasargs(fn, givenargs):
     args = []
     # only care about alias 'args', ignore 'args' set by extensions.wrapfunction
-    if not util.safehasattr(fn, b'_origfunc'):
+    if not hasattr(fn, '_origfunc'):
         args = getattr(fn, 'args', args)
     if args:
         cmd = b' '.join(map(procutil.shellquote, args))
@@ -575,7 +573,7 @@ def aliasinterpolate(name, args, cmd):
     return r.sub(lambda x: replacemap[x.group()], cmd)
 
 
-class cmdalias(object):
+class cmdalias:
     def __init__(self, ui, name, definition, cmdtable, source):
         self.name = self.cmd = name
         self.cmdname = b''
@@ -590,7 +588,7 @@ class cmdalias(object):
 
         try:
             aliases, entry = cmdutil.findcmd(self.name, cmdtable)
-            for alias, e in pycompat.iteritems(cmdtable):
+            for alias, e in cmdtable.items():
                 if e is entry:
                     self.cmd = alias
                     break
@@ -721,7 +719,7 @@ class cmdalias(object):
         }
         if name not in adefaults:
             raise AttributeError(name)
-        if self.badalias or util.safehasattr(self, b'shell'):
+        if self.badalias or hasattr(self, 'shell'):
             return adefaults[name]
         return getattr(self.fn, name)
 
@@ -747,7 +745,7 @@ class cmdalias(object):
             self.name,
             self.definition,
         )
-        if util.safehasattr(self, b'shell'):
+        if hasattr(self, 'shell'):
             return self.fn(ui, *args, **opts)
         else:
             try:
@@ -758,7 +756,7 @@ class cmdalias(object):
                 raise
 
 
-class lazyaliasentry(object):
+class lazyaliasentry:
     """like a typical command entry (func, opts, help), but is lazy"""
 
     def __init__(self, ui, name, definition, cmdtable, source):
@@ -838,9 +836,20 @@ def _parse(ui, args):
         cmd = None
         c = []
 
+    def global_opt_to_fancy_opt(opt_name):
+        # fancyopts() does this transform on `options`, but globalopts uses a
+        # '-', so that it is displayed in the help and accepted as input that
+        # way.
+        return opt_name.replace(b'-', b'_')
+
     # combine global options into local
     for o in commands.globalopts:
-        c.append((o[0], o[1], options[o[1]], o[3]))
+        name = global_opt_to_fancy_opt(o[1])
+
+        # The fancyopts name is needed for `options`, but the original name
+        # needs to be used in the second element here, or the parsing for the
+        # command verb fails, saying the command has no such option.
+        c.append((o[0], o[1], options[name], o[3]))
 
     try:
         args = fancyopts.fancyopts(args, c, cmdoptions, gnu=True)
@@ -849,7 +858,7 @@ def _parse(ui, args):
 
     # separate global options back out
     for o in commands.globalopts:
-        n = o[1]
+        n = global_opt_to_fancy_opt(o[1])
         options[n] = cmdoptions[n]
         del cmdoptions[n]
 
@@ -862,7 +871,7 @@ def _parseconfig(ui, config):
 
     for cfg in config:
         try:
-            name, value = [cfgelem.strip() for cfgelem in cfg.split(b'=', 1)]
+            name, value = (cfgelem.strip() for cfgelem in cfg.split(b'=', 1))
             section, name = name.split(b'.', 1)
             if not section or not name:
                 raise IndexError
@@ -880,6 +889,48 @@ def _parseconfig(ui, config):
     return configs
 
 
+def _parse_config_files(
+    ui, cmdargs: list[bytes], config_files: Iterable[bytes]
+) -> list[tuple[bytes, bytes, bytes, bytes]]:
+    """parse the --config-file options from the command line
+
+    A list of tuples containing (section, name, value, source) is returned,
+    in the order they were read.
+    """
+
+    configs: list[tuple[bytes, bytes, bytes, bytes]] = []
+
+    cfg = configmod.config()
+
+    for file in config_files:
+        try:
+            cfg.read(file)
+        except error.ConfigError as e:
+            raise error.InputError(
+                _(b'invalid --config-file content at %s') % e.location,
+                hint=e.message,
+            )
+        except FileNotFoundError:
+            hint = None
+            if b'--cwd' in cmdargs:
+                hint = _(b"this file is resolved before --cwd is processed")
+
+            raise error.InputError(
+                _(b'missing file "%s" for --config-file') % file, hint=hint
+            )
+
+    for section in cfg.sections():
+        for item in cfg.items(section):
+            name = item[0]
+            value = item[1]
+            src = cfg.source(section, name)
+
+            ui.setconfig(section, name, value, src)
+            configs.append((section, name, value, src))
+
+    return configs
+
+
 def _earlyparseopts(ui, args):
     options = {}
     fancyopts.fancyopts(
@@ -897,7 +948,13 @@ def _earlysplitopts(args):
     """Split args into a list of possible early options and remainder args"""
     shortoptions = b'R:'
     # TODO: perhaps 'debugger' should be included
-    longoptions = [b'cwd=', b'repository=', b'repo=', b'config=']
+    longoptions = [
+        b'cwd=',
+        b'repository=',
+        b'repo=',
+        b'config=',
+        b'config-file=',
+    ]
     return fancyopts.earlygetopt(
         args, shortoptions, longoptions, gnu=True, keepsep=True
     )
@@ -942,28 +999,17 @@ def runcommand(lui, repo, cmd, fullargs, ui, options, d, cmdpats, cmdoptions):
     return ret
 
 
-def _readsharedsourceconfig(ui, path):
-    """if the current repository is shared one, this tries to read
-    .hg/hgrc of shared source if we are in share-safe mode
+def _get_cwd() -> bytes:
+    """return the path to the current working directory
 
-    Config read is loaded into the ui object passed
-
-    This should be called before reading .hg/hgrc or the main repo
-    as that overrides config set in shared source"""
+    raise an Abort error in case of error.
+    """
     try:
-        with open(os.path.join(path, b".hg", b"requires"), "rb") as fp:
-            requirements = set(fp.read().splitlines())
-            if not (
-                requirementsmod.SHARESAFE_REQUIREMENT in requirements
-                and requirementsmod.SHARED_REQUIREMENT in requirements
-            ):
-                return
-            hgvfs = vfs.vfs(os.path.join(path, b".hg"))
-            sharedvfs = localrepo._getsharedvfs(hgvfs, requirements)
-            root = sharedvfs.base
-            ui.readconfig(sharedvfs.join(b"hgrc"), root)
-    except IOError:
-        pass
+        return encoding.getcwd()
+    except OSError as e:
+        msg = _(b"error getting current working directory: %s")
+        msg %= encoding.strtolocal(e.strerror)
+        raise error.Abort(msg)
 
 
 def _getlocal(ui, rpath, wd=None):
@@ -971,32 +1017,41 @@ def _getlocal(ui, rpath, wd=None):
 
     Takes paths in [cwd]/.hg/hgrc into account."
     """
-    if wd is None:
-        try:
-            wd = encoding.getcwd()
-        except OSError as e:
-            raise error.Abort(
-                _(b"error getting current working directory: %s")
-                % encoding.strtolocal(e.strerror)
-            )
+    cwd = _get_cwd()
+    # If using an alternate wd, temporarily switch to it so that relative
+    # paths are resolved correctly during config loading.
+    oldcwd = None
+    try:
+        if wd is None:
+            wd = cwd
+        else:
+            oldcwd = cwd
+            os.chdir(wd)
 
-    path = cmdutil.findrepo(wd) or b""
-    if not path:
-        lui = ui
-    else:
-        lui = ui.copy()
-        if rcutil.use_repo_hgrc():
-            _readsharedsourceconfig(lui, path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc"), path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc-not-shared"), path)
+        path = cmdutil.findrepo(wd) or b""
+        if not path:
+            lui = ui
+        else:
+            lui = ui.copy()
+            if rcutil.use_repo_hgrc():
+                for __, c_type, rc_path in rcutil.repo_components(path):
+                    assert c_type == b'path'
+                    lui.readconfig(rc_path, root=path)
 
-    if rpath:
-        path = urlutil.get_clone_path(lui, rpath)[0]
-        lui = ui.copy()
-        if rcutil.use_repo_hgrc():
-            _readsharedsourceconfig(lui, path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc"), path)
-            lui.readconfig(os.path.join(path, b".hg", b"hgrc-not-shared"), path)
+        if rpath:
+            # the specified path, might be defined in the [paths] section of
+            # the local repository. So we had to read the local config first
+            # even if it get overriden here.
+            path_obj = urlutil.get_clone_path_obj(lui, rpath)
+            path = path_obj.rawloc
+            lui = ui.copy()
+            if rcutil.use_repo_hgrc():
+                for __, c_type, rc_path in rcutil.repo_components(path):
+                    assert c_type == b'path'
+                    lui.readconfig(rc_path, root=path)
+    finally:
+        if oldcwd:
+            os.chdir(oldcwd)
 
     return path, lui
 
@@ -1025,7 +1080,7 @@ def _checkshellalias(lui, ui, args):
     cmd = aliases[0]
     fn = entry[0]
 
-    if cmd and util.safehasattr(fn, b'shell'):
+    if cmd and hasattr(fn, 'shell'):
         # shell alias shouldn't receive early options which are consumed by hg
         _earlyopts, args = _earlysplitopts(args)
         d = lambda: fn(ui, *args[1:])
@@ -1035,14 +1090,29 @@ def _checkshellalias(lui, ui, args):
 
 
 def _dispatch(req):
+    # check for cwd
+    old_cwd = None
+    cwd = req.earlyoptions[b'cwd']
+    try:
+        if cwd:
+            try:
+                old_cwd = encoding.getcwd()
+            except OSError as e:
+                raise error.Abort(
+                    _(b"error getting current working directory: %s")
+                    % encoding.strtolocal(e.strerror)
+                )
+            os.chdir(cwd)
+        return _dispatch_post_cwd(req)
+    finally:
+        if old_cwd is not None:
+            os.chdir(old_cwd)
+
+
+def _dispatch_post_cwd(req):
     args = req.args
     ui = req.ui
 
-    # check for cwd
-    cwd = req.earlyoptions[b'cwd']
-    if cwd:
-        os.chdir(cwd)
-
     rpath = req.earlyoptions[b'repository']
     path, lui = _getlocal(ui, rpath)
 
@@ -1104,13 +1174,32 @@ def _dispatch(req):
             encoding.fallbackencoding = fallback
 
         fullargs = args
-        cmd, func, args, options, cmdoptions = _parse(lui, args)
+        try:
+            cmd, func, args, options, cmdoptions = _parse(lui, args)
+        except error.CommandError as e:
+            cause = e.__context__
+            if isinstance(cause, getopt.GetoptError):
+                if cause.opt and "config".startswith(cause.opt):
+                    # pycompat._getoptbwrapper() decodes bytes with latin-1
+                    opt = cause.opt.encode('latin-1')
+                    all_long = {o[1] for o in commands.globalopts}
+                    possible = [o for o in all_long if o.startswith(opt)]
+
+                    if len(possible) != 1:
+                        raise error.InputError(
+                            _(b"option --config may not be abbreviated")
+                        )
+            raise
 
         # store the canonical command name in request object for later access
         req.canonical_command = cmd
 
         if options[b"config"] != req.earlyoptions[b"config"]:
             raise error.InputError(_(b"option --config may not be abbreviated"))
+        if options[b"config_file"] != req.earlyoptions[b"config_file"]:
+            raise error.InputError(
+                _(b"option --config-file may not be abbreviated")
+            )
         if options[b"cwd"] != req.earlyoptions[b"cwd"]:
             raise error.InputError(_(b"option --cwd may not be abbreviated"))
         if options[b"repository"] != req.earlyoptions[b"repository"]:
@@ -1210,7 +1299,7 @@ def _dispatch(req):
         assert func is not None  # help out pytype
         if not func.norepo:
             # use the repo from the request only if we don't have -R
-            if not rpath and not cwd:
+            if not rpath and not req.earlyoptions[b'cwd']:
                 repo = req.repo
 
             if repo:
@@ -1342,7 +1431,9 @@ def _exceptionwarning(ui):
     else:
         bugtracker = ui.config(b'ui', b'supportcontact')
         if bugtracker is None:
-            bugtracker = _(b"https://mercurial-scm.org/wiki/BugTracker")
+            bugtracker = _(
+                b"https://foss.heptapod.net/mercurial/mercurial-devel/-/issues"
+            )
         warning = (
             _(
                 b"** unknown exception encountered, "
