@@ -1,4 +1,3 @@
-# coding: utf8
 # ext-sidedata-4.py - small extension to test (differently still) the sidedata
 # logic
 #
@@ -9,7 +8,6 @@
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
-from __future__ import absolute_import
 
 from mercurial.revlogutils import sidedata
 
